from typing import Union

import numpy as np
from sklearn import datasets

from feltoken.node.config import Config


def load_data(config: Config) -> Union[tuple, str]:
    """Load data and return them in (X, y) formta."""
    if config.data == "test":
        # Demo data for testing
        X, y = datasets.load_diabetes(return_X_y=True)
        subset = np.random.choice(X.shape[0], 100, replace=False)
        return X[subset], y[subset]
    else:
        try:
            data = np.genfromtxt(config.data, delimiter=",")
            return data[:-1], data[-1]
        except Exception as e:
            raise Exception(f"Unable to load {config.data}\n{e}")



"""Model for performing federated averaging of models."""
from typing import Any

import numpy as np

ATTRIBUTE_LIST = ["coef_", "intercept_", "coefs_", "intercepts_"]


def _get_models_params(models: list[Any]) -> dict[str, list[np.ndarray]]:
    """Extract trainable parameters from scikit-learn models.

    Args:
        modesl: list of scikit-learn models.

    Returns:
        dictionary mapping attributes to list of values numpy arrays extracted from models.
    """
    params = {}
    for param in ATTRIBUTE_LIST:
        params[param] = []
        try:
            for model in models:
                params[param].append(getattr(model, param))
        except Exception:
            params.pop(param, None)

    return params


def _set_model_params(model: Any, params: dict[str, np.ndarray]):
    """Set new values of trainable params to scikit-learn models.

    Args:
        model: scikit-learn model.
        params: dictinary mapping attributes to numpy arrays.

    Returns:
        scikit-learn model with new values.
    """
    for param, value in params.items():
        setattr(model, param, value)
    return model


def average_models(models: list[Any]) -> Any:
    """Average trainable parameters of scikit-learn models.

    Args:
        models: list of scikit-learn models.

    Returns:
        scikit-learn model with new values.
    """
    params = _get_models_params(models)
    average_params = {}
    for param, values in params.items():
        val = np.mean(values, axis=0)
        average_params[param] = val.astype(values[0].dtype)

    model = _set_model_params(models[0], average_params)
    return model



import time
from optparse import Option
from pathlib import Path
from typing import Any, Optional, Union

import numpy as np
from eth_account.signers.local import LocalAccount as Account
from web3 import Web3
from web3.contract import Contract

from feltoken.core.average import average_models
from feltoken.core.contracts import to_dict
from feltoken.core.data import load_data
from feltoken.core.node import check_node_state, get_node, get_node_secret
from feltoken.core.storage import (
    export_model,
    ipfs_download_file,
    load_model,
    upload_encrypted_model,
    upload_final_model,
)
from feltoken.core.web3 import get_current_secret, get_project_contract, get_web3
from feltoken.node.config import Config, parse_args
from feltoken.node.training import train_model

# Path for saving logs and models during training
LOGS = Path(__file__).parent / "logs" / f"{time.time()}"


def _get_plan(project_contract: Contract) -> Optional[dict]:
    """Get latest running plan else return None."""
    if project_contract.functions.isPlanRunning().call():
        length = project_contract.functions.numPlans().call()
        plan = project_contract.functions.plans(length - 1).call()
        return to_dict(plan, "TrainingPlan")
    return None


def _watch_for_plan(project_contract: Contract):
    """Wait until new plan created."""
    # TODO: Use contract emiting events
    print("Waiting for a plan.")
    while True:
        plan = _get_plan(project_contract)
        if plan is not None:
            return plan
        time.sleep(3)


def _execute_rounds(
    data: Union[tuple, str],
    model: Any,
    plan: dict,
    plan_dir: Path,
    secret: bytes,
    account: Account,
    project_contract: Contract,
    w3: Web3,
    config: Config,
) -> Any:
    """Perform training rounds according to the training plan.

    Args:
        ...

    Returns:
        (object): scikit-learn model
    """
    num_rounds = plan["numRounds"]
    for i in range(num_rounds):
        print(f"\nRunning round {i}")
        round_dir = plan_dir / f"round_{i}"
        round_dir.mkdir(exist_ok=True)

        # 2. Execute training
        print("Training")
        model = train_model(model, data, config)

        # 3. Encrypt the model
        model_path = round_dir / "node_model.joblib"
        cid = upload_encrypted_model(model, model_path, secret)

        # 5. Send model to the contract (current round)
        tx = project_contract.functions.submitModel(cid).transact(
            {"from": account.address, "gasPrice": w3.eth.gas_price}
        )
        w3.eth.wait_for_transaction_receipt(tx)

        # 6. Download models and wait for round finished
        models = [model]
        downloaded = set()
        print("Waiting for other nodes to finish round.")
        while len(models) < plan["numNodes"]:
            length = project_contract.functions.getNodesLength().call()
            for node_idx in range(length):
                node = project_contract.functions.nodesArray(node_idx).call()
                node = to_dict(node, "Node")
                if (
                    node_idx in downloaded
                    or not node["activated"]
                    or node["_address"] == account.address
                ):
                    continue

                cid = project_contract.functions.getRoundModel(
                    i, node["_address"]
                ).call()
                if len(cid) < 5:
                    continue
                print(f"Downloading CID from node {node_idx}", cid)

                m_path = round_dir / f"model_node_{node_idx}.joblib"
                ipfs_download_file(cid, m_path, secret)
                models.append(load_model(m_path))
                downloaded.add(node_idx)

        print("Averaging models.", len(models))
        # 7. Average models
        model = average_models(models)
    return model


def task(data: Union[tuple, str], config: Config):
    """Connect to project contract and execute incoming training plans."""
    account = Web3().eth.account.from_key(config.account)
    w3 = get_web3(account, config.chain)
    print("Worker connected to chain id: ", w3.eth.chain_id)

    project_contract = get_project_contract(w3, config.contract)
    if not check_node_state(w3, project_contract, account):
        print("Script stoped.")
        return
    print("Node is ready for training.")

    # Obtain secret from the contract
    node = get_node(project_contract, account)
    SECRET = get_node_secret(project_contract, account)

    while True:
        plan = _watch_for_plan(project_contract)
        print("Executing a plan!")
        # Use random seed from contract
        np.random.seed(plan["randomSeed"])

        secret = get_current_secret(SECRET, node["entryKeyTurn"], plan["keyTurn"])

        # Creat directory for storing plan
        plan_index = project_contract.functions.numPlans().call()
        plan_dir = LOGS / f"plan_{plan_index}"
        plan_dir.mkdir(parents=True, exist_ok=True)

        # 1. Download model by CID
        base_model_path = plan_dir / "base_model.joblib"
        ipfs_download_file(plan["baseModelCID"], output_path=base_model_path)
        model = load_model(base_model_path)

        final_model = _execute_rounds(
            data,
            model,
            plan,
            plan_dir,
            secret,
            account,
            project_contract,
            w3,
            config,
        )
        print("Creating final model.")
        final_model_path = plan_dir / "final_model.joblib"
        export_model(model, final_model_path)

        # 8. Upload final model if coordinator
        if plan["finalNode"] == account.address:
            print("Node selected as a final one.")
            # Generate builder secret based on random seed and secret
            # So it is same for all nodes
            builder = project_contract.functions.builders(plan["creator"]).call()
            builder = to_dict(builder, "Builder")

            cid = upload_final_model(
                final_model, final_model_path, builder["publicKey"]
            )

            tx = project_contract.functions.finishPlan(cid).transact(
                {"from": account.address, "gasPrice": w3.eth.gas_price}
            )
            w3.eth.wait_for_transaction_receipt(tx)
            print("Final model uploaded and encrypted for a builder.")

        time.sleep(30)
        print("Plan finished!")


def main(args_str: Optional[str] = None):
    """Parse arguments and run worker task (watching contract and training models)."""
    config = parse_args(args_str)
    data = load_data(config)

    task(data, config)


if __name__ == "__main__":
    main()



"""Module for executing model training."""
import argparse
from typing import Any, Optional, Union, cast

import numpy as np

from feltoken.core.data import load_data
from feltoken.core.storage import export_model, load_model
from feltoken.node.config import Config


class TrainingConfig:
    model: str
    data: str
    output_model: str
    account: Optional[str]


# TODO: type for the model
def train_model(
    model: Any,
    data: Union[tuple[np.ndarray, np.ndarray], str],
    config: Union[TrainingConfig, Config],
) -> Any:
    """Universal model training function, it starts training depending on selected type.

    Args:
        model: initial model object with fit() method
        data: tuple[X, Y] or string depending on training type
        config: object specifing traing type

    Returns:
        new model object with trained weights
    """
    assert type(data) == tuple, "Invalid data type"
    model.fit(data[0], data[1])
    return model


def parse_args(args_str: Optional[str] = None) -> TrainingConfig:
    """Parse and partially validate arguments form command line.
    Arguments are parsed from string args_str or command line if args_str is None

    Args:
        args_str: string with arguments or None if using command line

    Returns:
        Parsed args object
    """
    parser = argparse.ArgumentParser(
        description="Script for training models, possible to execute from command line."
    )
    parser.add_argument(
        "--model",
        type=str,
        help="Path to model file which should be loaded and trained.",
    )
    parser.add_argument(
        "--data",
        type=str,
        default="test",
        help="Path to CSV file with data. Last column is considered as Y.",
    )
    parser.add_argument(
        "--output_model",
        type=str,
        help="Path to store the final model file.",
    )
    args = parser.parse_args(args_str)

    return cast(TrainingConfig, args)


def main():
    """Main function for exectuting from command line."""
    args = parse_args()
    model = load_model(args.model)

    data = load_data(args)

    model = train_model(model, data, args)
    export_model(model, args.output_model)


if __name__ == "__main__":
    main()



"""Module for storing and managing data files at IPFS/Filecoin using web3.storage."""
import os
import time
from io import BytesIO
from pathlib import Path
from typing import Any, BinaryIO, Optional, Union

import httpx
import joblib
from httpx import Response

from feltoken.core.web3 import decrypt_bytes, encrypt_bytes, encrypt_nacl

# File type definiton
FileType = Union[str, Path, BinaryIO]
PathType = Union[str, Path]


def load_model(filename: FileType) -> Any:
    """Abstraction function for loading models.

    Args:
        filename: filepath or file-like object to load

    Returns:
        model object
    """
    return joblib.load(filename)


def export_model(model: Any, path: PathType):
    """Abstraction function for exporting model to file."""
    joblib.dump(model, path)


def model_to_bytes(model: Any, path: PathType = "/tmp/model") -> bytes:
    """Convert model to bytes which can be stored/exchanged/loaded.

    Args:
        model: model object
        path: path-like object where to store the model

    Return:
        bytes representing the model
    """
    # TODO: Optimize this part, so we don't need to r/w, json.dump to memory,
    #       But sometimes we actually want to store the model so it's ok so far
    export_model(model, path)
    with open(path, "rb") as f:
        return f.read()


def ipfs_upload_file(file: BinaryIO) -> Optional[Response]:
    """Upload file to IPFS using web3.storage.

    Args:
        file: file-like object in byte mode.

    Returns:
        Response: httpx response object
    """
    # TODO: Throw some error when upload fails on all 3 times
    for _ in range(3):
        try:
            return httpx.post(
                "https://api.web3.storage/upload",
                headers={"Authorization": "Bearer " + os.environ["WEB3_STORAGE_TOKEN"]},
                files={"file": file},
                timeout=None,
            )
        except (httpx.ReadTimeout, httpx.ConnectError):
            print("Upload error - retry")
            time.sleep(5)


def ipfs_download_file(
    cid: str, output_path: Optional[PathType] = None, secret: Optional[bytes] = None
) -> bytes:
    """Download file stored in IPFS.

    Args:
        cid: string describing location of the file.
        output_path: if set file will be stored at this path.

    Returns:
        Response content: httpx response content object
    """
    # TODO: Throw some error when no res object
    for _ in range(5):
        try:
            res = httpx.get(f"https://ipfs.io/ipfs/{cid}", timeout=30.0)
        except (httpx.ReadTimeout, httpx.ConnectError):
            print("Connection timeout - retry")
            time.sleep(5)

    content = res.content
    if secret is not None:
        content = decrypt_bytes(res.content, secret)

    if output_path is not None:
        with open(output_path, "wb") as f:
            f.write(content)

    return content


def upload_final_model(model: Any, model_path: PathType, builder_key: bytes) -> str:
    """Encrypt and upload final model for builder to IPFS.

    Args:
        model: scikit-learn trained model
        model_path: path to save final model
        builder_key: public builder key to use for encryption

    Returns:
        (str): CID of uploaded file.
    """
    model_bytes = model_to_bytes(model, model_path)
    # TODO: Right now only builder will be able to decrypt model
    #       Maybe upload extra finall model for nodes??
    #       Or use secret such that all nodes can calculate it (emph key?)
    encrypted_model = encrypt_nacl(builder_key, model_bytes)

    # 4. Upload file to IPFS
    res = ipfs_upload_file(BytesIO(encrypted_model))
    return res.json()["cid"]


def upload_encrypted_model(model: Any, model_path: PathType, secret: bytes) -> str:
    """Encrypt and upload model to IPFS.

    Args:
        model: scikit-learn trained model
        model_path: path to save final model
        secret: secret key for encryption

    Returns:
        (str): CID of uploaded file.
    """
    model_bytes = model_to_bytes(model, model_path)
    encrypted_model = encrypt_bytes(model_bytes, secret)

    # 4. Upload file to IPFS
    res = ipfs_upload_file(BytesIO(encrypted_model))
    return res.json()["cid"]



"""Module providing helpful functions for working with contracts."""

# Mapping for training Plan struct
PI = {
    "creator": 0,
    "finalNode": 1,
    "randomSeed": 2,
    "baseModelCID": 3,
    "finalModelCID": 4,
    "numRounds": 5,
    "numNodes": 6,
    "totalReward": 7,
    "nodeReward": 8,
    "keyTurn": 9,
}


# Mapping for Node struct
NI = {
    "_address": 0,
    "activated": 1,
    "entryKeyTurn": 2,
}

# Mapping for Builder struct
BI = {
    "_address": 0,
    "publicKey": 1,
}


structures = {
    "TrainingPlan": PI,
    "Node": NI,
    "Builder": BI,
}


def to_dict(values: list, struct_name: str) -> dict:
    """Map list to dictionary based on the structure from contract."""
    d = {}
    mapping = structures[struct_name]
    for k, idx in mapping.items():
        d[k] = values[idx]
    return d



"""FELToken python package intended for running data provider code.

This code connects to the specified smart contract and trains the specified
models on provided data.

Entry command:

```bash
felt-node-worker --chain <80001> --contract <address> --account main --data <data_path.csv>
```
"""
import re
from pathlib import Path
from urllib import request

from setuptools import find_packages, setup

PATH = Path(__file__).parent.absolute()
DOCLINES = (__doc__ or "").split("\n")

CLASSIFIERS = """\
Intended Audience :: Science/Research
Intended Audience :: Developers
Programming Language :: Python
Programming Language :: Python :: 3
Programming Language :: Python :: 3.8
Programming Language :: Python :: 3.9
Programming Language :: Python :: 3.10
Programming Language :: Python :: 3 :: Only
Topic :: Scientific/Engineering
"""

# Download project contract artifacts
artifacts = PATH / "feltoken/artifacts"
artifacts.mkdir(parents=True, exist_ok=True)

# TODO: Add supported chains/remove contracts
remote_url = "https://raw.githubusercontent.com/FELToken/smart-contracts/main/build/deployments/80001/ProjectContract.json"
request.urlretrieve(remote_url, artifacts / "ProjectContract.json")


def parse_requirements(file_name):
    """
    from:
        http://cburgmer.posterous.com/pip-requirementstxt-and-setuppy
    """
    requirements = []
    with open(file_name, "r") as f:
        for line in f:
            if re.match(r"(\s*#)|(\s*$)", line):
                continue
            if re.match(r"\s*-e\s+", line):
                requirements.append(
                    re.sub(r"\s*-e\s+.*#egg=(.*)$", r"\1", line).strip()
                )
            elif re.match(r"\s*-f\s+", line):
                pass
            else:
                requirements.append(line.strip())
    return requirements


requirements = parse_requirements(PATH / "requirements.txt")


setup(
    name="feltoken",
    version="0.2.1",
    packages=find_packages(),
    maintainer="FELToken",
    maintainer_email="info@bretahajek.com",
    description=DOCLINES[0],
    long_description="\n".join(DOCLINES[2:]),
    long_description_content_type="text/markdown",
    keywords=["Federated Learning", "Web3", "Machine Learning"],
    url="https://feltoken.ai/",
    author="FELToken",
    project_urls={
        "Bug Tracker": "https://github.com/FELToken/feltoken.py/issues",
        "Documentation": "https://docs.feltoken.ai/",
        "Source Code": "https://github.com/FELToken/feltoken.py",
    },
    license="GPL-3.0 License",
    classifiers=[_f for _f in CLASSIFIERS.split("\n") if _f],
    platforms=["Windows", "Linux", "Solaris", "Mac OS-X", "Unix"],
    python_requires=">=3.8",
    install_requires=requirements,
    zip_safe=False,
    entry_points={
        "console_scripts": [
            "felt-node-worker = feltoken.node.background_worker:main",
            "feltoken-train = feltoken.node.training:main",
        ],
    },
    package_data={
        "feltoken": ["artifacts/*.json", "artifacts/contracts/*.json"],
    },
)



"""Module supporting work of the builder."""
from pathlib import Path

import joblib

from feltoken.core.storage import ipfs_upload_file


def upload_model(model: Any) -> str:
    """Function for uploading scikit-learn model to ipfs."""
    tmp_path = Path("tmp_model.joblib")
    joblib.dump(model, tmp_path)
    with open(tmp_path, "rb") as f:
        res = ipfs_upload_file(f)
    tmp_path.unlink()
    return res.json()["cid"]



import time

from eth_account.signers.local import LocalAccount as Account
from web3 import Web3
from web3.contract import Contract

from feltoken.core.contracts import to_dict
from feltoken.core.prompts import yes_no_prompt
from feltoken.core.web3 import decrypt_nacl, export_public_key


def get_node(project_contract: Contract, account: Account) -> dict:
    """Get node data for given account."""
    index = project_contract.functions.nodeState(account.address).call()
    assert (
        index >= 3
    ), f"Node with this address ({account.address}) isn't approved by contract."

    node = project_contract.functions.nodesArray(index - 3).call()
    return to_dict(node, "Node")


def get_node_secret(project_contract: Contract, account: Account) -> bytes:
    """Get shared secret for node represented by account."""
    secret = b"".join(project_contract.functions.getNodeSecret(account.address).call())
    return decrypt_nacl(account.key, secret)


def _check_node_isactive(
    w3: Web3, project_contract: Contract, account: Account
) -> bool:
    """Check if accepted node has active status, activate if user allows it."""
    node = get_node(project_contract, account)
    if not node["activated"]:
        print("Node with this account is set as inactive.")
        if yes_no_prompt("Do you want to activate the node?", default=False):
            # TODO: Add this once contract updated
            tx = project_contract.functions.activate().transact(
                {"from": account.address, "gasPrice": w3.eth.gas_price},
            )
            w3.eth.wait_for_transaction_receipt(tx)
            print("Node activated.")
        else:
            print("Node remains inactive.")
            return False

    return True


def check_node_state(w3: Web3, project_contract: Contract, account: Account) -> bool:
    """Check node state depending on a state perform action:

    State:
        0 - no access request - ask if user wants to join:
                                yes - perform join, got to state 1
                                no - return False
        1 - pending - notify about pending application, wait for state update
        2 - declined - return False
        3+ - accepted - return True

    Args:
        project_contract: project contract instance
        account: web3 account instance

    Returns:
        boolean - True if account can participate as data provider in project
    """
    while True:
        index = project_contract.functions.nodeState(account.address).call()
        if index == 0:
            # Request join as data provider
            print("You haven't requested access to project yet.")
            if yes_no_prompt("Do you want to join the project?", default=False):
                public_key = export_public_key(account.key)
                tx = project_contract.functions.requestJoinNode(public_key).transact(
                    {"from": account.address, "gasPrice": w3.eth.gas_price},
                )
                w3.eth.wait_for_transaction_receipt(tx)
                print("Request to join sent. Waiting to be accepted.")
                continue
            else:
                print("You decided not to join project.")
                return False

        elif index == 1:
            print("Waiting to be accepted.")
            time.sleep(10)
            continue

        elif index == 2:
            print("Node was declined from participation in this project.")
            return False

        # Check if node is set to active
        return _check_node_isactive(w3, project_contract, account)



"""Module for managing connection to chain and contracts."""
import hashlib
import json
from base64 import a85decode, a85encode
from pathlib import Path
from urllib import request

# TODO: Replace the encryption functions with something else?
from ecies.utils import aes_decrypt, aes_encrypt
from eth_account.signers.local import LocalAccount as Account
from eth_typing.evm import Address
from nacl.public import Box, PrivateKey, PublicKey
from web3 import Web3
from web3.contract import Contract
from web3.gas_strategies.time_based import medium_gas_price_strategy
from web3.middleware import construct_sign_and_send_raw_middleware

BUILD_FOLDER = Path(__file__).parent.parent / "artifacts"

DEPLOYMENTS_GIT = (
    "https://raw.githubusercontent.com/FELToken/smart-contracts/main/build/deployments"
)

CHAIN_ID_MAP = {
    1337: "http://127.0.0.1:8545",
    80001: "https://rpc-mumbai.maticvigil.com/v1/b5019eba066c735d9bd10b800ca694ed720e8d87",
}


def get_web3(account: Account, chain_id: int) -> Web3:
    """Get connection to web3."""
    w3 = Web3(Web3.HTTPProvider(CHAIN_ID_MAP[chain_id]))
    w3.eth.set_gas_price_strategy(medium_gas_price_strategy)
    w3.middleware_onion.add(construct_sign_and_send_raw_middleware(account))
    w3.eth.default_account = account.address
    return w3


def _download_abi(contract: str):
    """Download contract ABI file based on contract name (if missing)."""
    if not (BUILD_FOLDER / f"{contract}.json").exists():
        BUILD_FOLDER.mkdir(parents=True, exist_ok=True)
        remote_url = f"{DEPLOYMENTS_GIT}/{contract}.json"
        request.urlretrieve(remote_url, BUILD_FOLDER / f"{contract}.json")


def get_project_contract(w3: Web3, address: Address) -> Contract:
    """Load project contract on current chain from build folder."""
    _download_abi("ProjectContract")
    contract = json.load((BUILD_FOLDER / "ProjectContract.json").open())
    return w3.eth.contract(address=address, abi=contract["abi"])


def export_public_key(private_key: bytes) -> bytes:
    """Export public key for contract join request.

    Args:
        private_key: bytes representing private key

    Returns:
        32 bytes representing public key
    """
    return bytes(PrivateKey(private_key).public_key)


def encrypt_nacl(public_key: bytes, data: bytes) -> bytes:
    """Encryption function using NaCl box compatible with MetaMask
    For implementation used in MetaMask look into: https://github.com/MetaMask/eth-sig-util

    Args:
        public_key: public key of recipient
        data: message data

    Returns:
        encrypted data
    """
    emph_key = PrivateKey.generate()
    enc_box = Box(emph_key, PublicKey(public_key))
    # Encryption is required to work with MetaMask decryption (requires utf8)
    data = a85encode(data)
    ciphertext = enc_box.encrypt(data)
    return bytes(emph_key.public_key) + ciphertext


def decrypt_nacl(private_key: bytes, data: bytes) -> bytes:
    """Decryption function using NaCl box compatible with MetaMask
    For implementation used in MetaMask look into: https://github.com/MetaMask/eth-sig-util

    Args:
        private_key: private key to decrypt with
        data: encrypted message data

    Returns:
        decrypted data
    """
    emph_key, ciphertext = data[:32], data[32:]
    box = Box(PrivateKey(private_key), PublicKey(emph_key))
    return a85decode(box.decrypt(ciphertext))


def encrypt_bytes(bytes: bytes, secret: bytes) -> bytes:
    """Encrypt bytes (model) for storing in contract/IPFS."""
    return aes_encrypt(secret, bytes)


def decrypt_bytes(ciphertext: bytes, secret: bytes) -> bytes:
    """Decrypt bytes (model) stored in contract/IPFS."""
    return aes_decrypt(secret, ciphertext)


def get_current_secret(secret: bytes, entry_key_turn: int, key_turn: int) -> bytes:
    """Calculate shared secret at current state."""
    for _ in range(entry_key_turn, key_turn):
        secret = hashlib.sha256(secret).digest()
    return secret



"""Module for loading data provider (Node) config."""
import argparse
import os
from getpass import getpass
from typing import Optional, cast

from dotenv import load_dotenv

# Load dotenv at the beginning of the program
load_dotenv()


KEYS = {
    "main": os.getenv("PRIVATE_KEY"),
    "node1": os.getenv("NODE1_PRIVATE_KEY"),
    "node2": os.getenv("NODE2_PRIVATE_KEY"),
}


class Config:
    """Config class type."""

    chain: int
    contract: str
    account: str
    data: str
    output_model: str


def _verify_config(config: Config) -> Config:
    """Verify if provided values are correct."""
    assert config.chain in [
        1337,
        80001,
        137,
    ], "Invalid chain id or chain id is not supported (suppoerted: 1337, 137, 80001)"
    assert len(config.contract) == 42, "The contract address has invalid length."

    if config.account in KEYS and KEYS[config.account] is not None:
        config.account = KEYS[config.account]
    else:
        config.account = getpass(
            "Please provide your private key (exported from MetaMask):"
        )

    # Get web3 token if missing
    if "WEB3_STORAGE_TOKEN" not in os.environ or not os.getenv("WEB3_STORAGE_TOKEN"):
        os.environ["WEB3_STORAGE_TOKEN"] = getpass(
            "Please input your web3.storage API token:"
        )

    return config


def parse_args(args_str: Optional[str] = None) -> Config:
    """Parse and partially validate arguments form command line.
    Arguments are parsed from string args_str or command line if args_str is None

    Args:
        args_str: string with arguments or None if using command line

    Returns:
        Parsed args object
    """
    parser = argparse.ArgumentParser(
        description="Data provider worker script managing the trainig."
    )
    parser.add_argument(
        "--chain",
        type=int,
        help="Chain Id of chain to which should be the worker connected.",
    )
    parser.add_argument("--contract", type=str, help="Contract address")
    parser.add_argument(
        "--account",
        type=str,
        default="main",
        help="Name of account to use as specified in .env (main, node1, node2)",
    )
    parser.add_argument(
        "--data",
        type=str,
        default="test",
        help=("Path to CSV file with data. Last column is considered as Y."),
    )
    args = parser.parse_args(args_str)

    return _verify_config(cast(Config, args))
